#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a GLib plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; GIO Objects
;;; package: (glib gio)

(C-include "glib")

(define (open-input-gfile uri)
  (let* ((uri* (->uri* uri 'open-input-gfile))
	 (gstream
	  (with-glib-lock
	   (lambda ()
	     (let* ((gfile (make-gfile uri*))
		    (gstream (gfile-read gfile)))
	       (gobject-unref! gfile)
	       gstream))))
	 (port (make-generic-i/o-port (make-binary-port
				       (make-g-stream-source gstream)
				       #f
				       'open-input-gfile)
				      (default-object)
				      'open-input-gfile)))
    ;;(port/set-coding port 'iso-8859-1)
    ;;(port/set-line-ending port 'newline)
    port))

(define (->uri* object caller)
  (let ((uri* (->uri object caller)))
    (if (uri-absolute? uri*)
	uri*
	(merge-uris (->simple-namestring object)
		    "file:///"))))

(define-integrable (->simple-namestring pathname)
  (->namestring (->simple-pathname pathname)))

(define (->simple-pathname pathname)
  (let loop ((simpler (pathname-simplify
		       (merge-pathnames pathname (working-directory-pathname))))
	     (count 1))
    (let ((again (pathname-simplify simpler)))
      (cond ((fix:> count 100) (error "Could not simplify:" pathname))
	    ((pathname=? again simpler) again)
	    (else (loop again (fix:1+ count)))))))

(define (make-g-stream-source gstream)
  (let ((open? #t))
    (make-non-channel-input-source
      (named-lambda (g-stream-source/has-bytes?)
	#t)
      (named-lambda (g-stream-source/read-bytes! buffer start end)
	(with-glib-lock
	 (lambda () (g-input-stream-read gstream buffer start end))))
      (named-lambda (g-stream-source/close)
	(if open?
	    (let ((value (with-glib-lock
			  (lambda () (g-input-stream-close gstream)))))
	      (set! open? #f)
	      value))))))

(define (open-output-gfile uri)
  (let* ((uri* (->uri* uri 'open-output-gfile))
	 (gstream
	  (with-glib-lock
	   (lambda ()
	     (let* ((gfile (make-gfile uri*))
		    (gstream (gfile-replace gfile #f #t 'private)))
	       (gobject-unref! gfile)
	       gstream))))
	 (port (make-generic-i/o-port (make-binary-port
				       #f
				       (make-g-stream-sink gstream)
				       'open-output-gfile)
				      (default-object)
				      'open-output-gfile)))
    ;;(port/set-coding port 'iso-8859-1)
    ;;(port/set-line-ending port 'newline)
    port))

(define (make-g-stream-sink gstream)
  (let ((open? #t))
    (make-non-channel-output-sink
     (named-lambda (g-stream-sink/write-bytes buffer start end)
       (with-glib-lock
	(lambda () (g-output-stream-write gstream buffer start end))))
     (named-lambda (g-stream-sink/close)
       (if open?
	   (let ((value (with-glib-lock
			 (lambda () (g-output-stream-close gstream)))))
	     (set! open? #f)
	     value))))))

(define (gdirectory-read uri)
  (let ((uri* (->uri* uri 'gdirectory-read)))
    (with-glib-lock
     (lambda ()
       (let ((gfile (make-gfile uri*)))
	 (map! (lambda (ginfo)
		 (let ((name (gfile-info-get-attribute-value
			      ginfo "standard::name")))
		   (gobject-unref! ginfo)
		   name))
	       (let ((children (gfile-children gfile "standard::name")))
		 (gobject-unref! gfile)
		 children)))))))

(define (gfile-children gfile attributes)
  (let ((genum (gfile-enumerate-children gfile attributes #f)))
    (let loop ((ginfos '()))
      (let ((next (gfile-enumerator-next-files genum 100)))
	(if (null? next)
	    (begin
	      (gfile-enumerator-close genum)
	      (gobject-unref! genum)
	      ginfos)
	    (loop (append! ginfos next)))))))

(define-class <gio>
    (<gobject>)

  (priority
   define standard initial-value 10)

  (queue
   define accessor initializer (lambda () (make-thread-queue 1)))

  (gio-cleanup-info
   define accessor accessor gio-cleanup-info initializer
   (lambda ()
     (make-gio-cleanup-info
      #f #f (make-gcancellable) (make-alien '(* |GError|))))))

;;; When a <gio> is GCed, any pending operation must be cancelled.
;;; This ensures that the operation's finish callback will not be
;;; called and can be safely de-registered.  The gio-cleanup-info
;;; includes the GCancellable, the finish callback id, AND a flag to
;;; indicate whether an operation is pending and thus whether the
;;; GCancellable should be used.  It also includes the GError pointer
;;; which, if not null, references a GError that must be freed.

(define-structure gio-cleanup-info
  pending-op	     ; #f, <opname>, closed or error.  The first one
		     ; means "idle" and the last two are more
		     ; permanent states than "op"s.  <opname> might be
		     ; open, read, skip, write, query, next, close,
		     ; etc.
  callback-id	     ; #f or op's finish callback ID
  gcancellable	     ; a GCancellable alien
  gerror-pointer)    ; a malloced (* GError) alien

(define-method initialize-instance ((object <gio>))
  (call-next-method object)
  (let* ((gio-info (gio-cleanup-info object))
	 (gerror* (gio-cleanup-info-gerror-pointer gio-info)))
    (assert-glib-locked '(initialize-instance <gio>))
    (C-call "g_try_malloc0" gerror* (C-sizeof "* GError"))
    (error-if-null gerror* "Could not create:" gerror*)))

(define-syntax cleanup-callback-id
  (sc-macro-transformer
   (lambda (form environment)
     (if (syntax-match? '(identifier symbol symbol) (cdr form))
	 (let ((info (close-syntax (cadr form) environment))
	       (type-name (caddr form))
	       (slot (cadddr form)))
       (let ((accessor (close-syntax
			(symbol type-name '-cleanup-info- slot)
			environment))
	     (modifier (close-syntax
			(symbol 'set- type-name '-cleanup-info- slot '!)
			environment)))
	 `(let ((id (,accessor ,info)))
	       (if id
		   (begin
		    (de-register-c-callback id)
		    (,modifier ,info #f))))))))))

(define-integrable-operator (cleanup-gerror-pointer gerror*)
  (assert-glib-locked 'cleanup-gerror-pointer)
  (if (not (alien-null? gerror*))
      (let ((gerror (make-alien '|GError|)))
	(C-> gerror* "* GError" gerror)
	(if (not (alien-null? gerror))
	    (C-call "g_error_free" gerror))
	(C-call "g_free" gerror*)
	(alien-null! gerror*))))

(define-integrable-operator (cleanup-gio gio-info)
  (assert-glib-locked 'cleanup-gio)
  (let ((pending-op (gio-cleanup-info-pending-op gio-info)))
    (if (not (memq pending-op '(#f error closed)))
	(C-call "g_cancellable_cancel"
		(gobject-alien (gio-cleanup-info-gcancellable gio-info)))))
  (cleanup-callback-id gio-info gio callback-id)
  (gobject-unref! (gio-cleanup-info-gcancellable gio-info))
  (cleanup-gerror-pointer (gio-cleanup-info-gerror-pointer gio-info)))

(define-integrable (guarantee-gio-idle gio)
  (let* ((gio-info (gio-cleanup-info gio))
	 (pending-op (gio-cleanup-info-pending-op gio-info)))
    (if (not (memq pending-op '(#f error closed)))
	(error "Operation pending:" gio))
    (if pending-op
	(error "Not open:" gio))))

(define-class <g-stream>
    (<gio>)
 (buffer define standard initializer (lambda () (malloc buffer-size 'uchar)))
 (buffer-size define standard initializer (lambda () buffer-size)))

(define buffer-size #x1000)

(define-class (<g-input-stream> (constructor ()))
    (<g-stream>)
  (cleanup-info
   define accessor initializer (lambda ()
				 (make-g-input-stream-cleanup-info #f #f))))

(define-structure g-input-stream-cleanup-info
  ;; To avoid registering a read or skip finish callback for every
  ;; read or skip (a LOT of de/registering!), the open operation
  ;; (i.e. gfile-read) registers them in advance.
  read-id			   ; #f or the read finish callback ID
  skip-id)			   ; #f or the skip finish callback ID

(define-method initialize-instance ((object <g-input-stream>))
  (call-next-method object)
  (add-glib-cleanup object (make-g-input-stream-cleanup
			    (gio-cleanup-info object)
			    (g-input-stream-cleanup-info object))))

(define (make-g-input-stream-cleanup gio-info info)
  (named-lambda (g-input-stream-cleanup)
    (cleanup-g-input-stream gio-info info)))

(define (cleanup-g-input-stream gio-info info)
  (assert-without-interruption 'cleanup-g-input-stream)
  (assert-glib-locked 'cleanup-g-input-stream)
  (cleanup-gio gio-info)
  (cleanup-callback-id info g-input-stream read-id)
  (cleanup-callback-id info g-input-stream skip-id))

(define-integrable (dequeue! queue)
  (without-glib-lock
   (lambda ()
     (thread-queue/dequeue! queue))))

(define (g-input-stream-read gstream buffer start end)
  (assert-glib-locked 'g-input-stream-read)
  (let* ((gio-info (gio-cleanup-info gstream))
	 (info (g-input-stream-cleanup-info gstream))
	 (callback-id (g-input-stream-cleanup-info-read-id info)))
    (guarantee-gio-idle gstream)
    (let* ((count (fix:- end start))
	   (async-buffer (ensure-buffer gstream count)))
      (set-gio-cleanup-info-pending-op! gio-info 'read)
      (C-call "g_input_stream_read_async"
	      (gobject-alien gstream)
	      async-buffer
	      count
	      (gio-priority gstream)
	      (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	      (C-callback "async_ready")
	      callback-id)
      (let* ((queue (gio-queue gstream))
	     (value (dequeue! queue)))
	(if (string? value)
	    (begin
	      (set-gio-cleanup-info-pending-op! gio-info 'error)
	      (error "Error reading:" gstream value))
	    (begin
	      (c-peek-bytes async-buffer 0 value buffer start)
	      (set-gio-cleanup-info-pending-op! gio-info #f)
	      value))))))

(define-integrable (ensure-buffer gstream count)
  (let ((size (g-stream-buffer-size gstream))
	(buffer (g-stream-buffer gstream)))
    (if (fix:< size count)
	(begin
	  (free buffer)
	  (let ((buffer* (malloc count 'uchar)))
	    (set-g-stream-buffer! gstream buffer*)
	    (set-g-stream-buffer-size! gstream count)
	    buffer*))
	buffer)))

(define (make-read-finish-callback queue gerror*)
  (C-callback
   (named-lambda (read-finish-callback source result)
     (assert-glib-locked 'read-finish-callback)
     (let ((bytes (C-call "g_input_stream_read_finish" source result gerror*)))
       (if (fix:= bytes -1)
	   (let ((message (%gerror-message gerror*)))
	     (%trace ";read-finish-callback "message" "queue"\n")
	     (%queue! queue message))
	   (begin
	     (%trace ";read-finish-callback "bytes" "queue"\n")
	     (%queue! queue bytes)))))))

(define-integrable-operator (%gerror-message pointer)
  (let* ((gerror (C-> pointer "* GError"))
	 (message (or (and (not (alien-null? gerror))
			   (c-peek-cstring (C-> gerror "GError message")))
		      "GError pointer not set.")))
    (if (not (alien-null? gerror))
	(begin
	  (C->= pointer "* GError" 0)
	  (C-call "g_error_free" gerror)))
    message))

(define-integrable (%queue! queue value)
  ;; The GIO finish callbacks use this procedure to queue a value on a
  ;; g-stream's queue AND signal the main loop if Scheme has become
  ;; runnable.
  (thread-queue/queue! queue value)
  (maybe-yield-glib))

(define (g-input-stream-skip gstream count)
  (assert-glib-locked 'g-input-stream-skip)
  (let* ((gio-info (gio-cleanup-info gstream))
	 (info (g-input-stream-cleanup-info gstream))
	 (callback-id (g-input-stream-cleanup-info-skip-id info)))
    (guarantee-gio-idle gstream)
    (set-gio-cleanup-info-pending-op! gio-info 'skip)
    (C-call "g_input_stream_skip_async"
	    (gobject-alien gstream)
	    count
	    (gio-priority gstream)
	    (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	    (C-callback "async_ready")
	    callback-id)
    (let* ((queue (gio-queue gstream))
	   (value (dequeue! queue)))
      (if (string? value)
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info 'error)
	    (error "Error skipping:" gstream value))
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info #f)
	    value)))))

(define (make-skip-finish-callback queue gerror*)
  (C-callback
   (named-lambda (skip-finish-callback source result)
     (assert-glib-locked 'skip-finish-callback)
     (let ((bytes (C-call "g_input_stream_skip_finish" source result gerror*)))
       (if (fix:= bytes -1)
	   (let ((message (%gerror-message gerror*)))
	     (%trace ";skip-finish-callback "message" "queue"\n")
	     (%queue! queue message))
	   (begin
	     (%trace ";skip-finish-callback "bytes" "queue"\n")
	     (%queue! queue bytes)))))))

(define (g-input-stream-close gstream)
  (assert-glib-locked 'g-input-stream-close)
  (gfile-close gstream
	       (named-lambda (close-input
			      gstream* priority gcancellable* callback id)
		 (C-call "g_input_stream_close_async"
			 gstream* priority gcancellable* callback id))
	       make-input-close-finish-callback
	       (named-lambda (close-input-cleanup gio-info)
		 (free (g-stream-buffer gstream))
		 (cleanup-g-input-stream
		  gio-info (g-input-stream-cleanup-info gstream)))))

(define (gfile-close gio callout make-callback cleanup)
  (assert-glib-locked 'gfile-close)
  (let* ((gio-info (gio-cleanup-info gio))
	 (queue (gio-queue gio))
	 (gerror* (gio-cleanup-info-gerror-pointer gio-info)))
    (guarantee-gio-idle gio)
    (let ((callback-id
	   (without-interruption	;don't leak callback IDs
	    (lambda ()
	      (let ((old (gio-cleanup-info-callback-id gio-info)))
		(if old (de-register-c-callback old)))
	      (let ((id (make-callback queue gerror*)))
		(set-gio-cleanup-info-pending-op! gio-info 'close)
		(set-gio-cleanup-info-callback-id! gio-info id)
		id)))))
      (callout (gobject-alien gio)
	       (gio-priority gio)
	       (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	       (C-callback "async_ready")
	       callback-id)
      (let ((value (dequeue! queue)))
	(cond ((eq? value #t)
	       (without-interruption
		(lambda ()
		  (set-gio-cleanup-info-pending-op! gio-info 'closed)
		  (cleanup gio-info)))
	       unspecific)
	      ((string? value)
	       (set-gio-cleanup-info-pending-op! gio-info 'error)
	       (error "Error during close:" gio value))
	      (else
	       (error "Unexpected:" value gio)))))))

(define (make-input-close-finish-callback queue gerror*)
  (C-callback
   (named-lambda (input-close-finish-callback source result)
     (assert-glib-locked 'input-close-finish-callback)
     (if (fix:zero?
	  (C-call "g_input_stream_close_finish" source result gerror*))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";input-close-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";input-close-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define-class (<g-output-stream> (constructor ()))
    (<g-stream>)
  (cleanup-info
   define accessor initializer (lambda ()
				 (make-g-output-stream-cleanup-info #f #f))))

(define-structure g-output-stream-cleanup-info
  ;; To avoid registering a write or flush finish callback for every
  ;; write or flush (a LOT of de/registering!), the open operation
  ;; (i.e. gfile-replace) registers them in advance.
  write-id			  ; #f or the write finish callback ID
  flush-id)			  ; #f or the flush finish callback ID

(define-method initialize-instance ((object <g-output-stream>))
  (call-next-method object)
  (add-glib-cleanup object (make-g-output-stream-cleanup
			    (gio-cleanup-info object)
			    (g-output-stream-cleanup-info object))))

(define (make-g-output-stream-cleanup gio-info info)
  (named-lambda (g-output-stream-cleanup)
    (cleanup-g-output-stream gio-info info)))

(define (cleanup-g-output-stream gio-info info)
  (assert-without-interruption 'cleanup-g-output-stream)
  (assert-glib-locked 'cleanup-g-output-stream)
  (cleanup-gio gio-info)
  (cleanup-callback-id info g-output-stream write-id)
  (cleanup-callback-id info g-output-stream flush-id))

(define (g-output-stream-write gstream buffer start end)
  (assert-glib-locked 'g-output-stream-write)
  (let* ((gio-info (gio-cleanup-info gstream))
	 (info (g-output-stream-cleanup-info gstream))
	 (callback-id (g-output-stream-cleanup-info-write-id info)))
    (guarantee-gio-idle gstream)
    (let* ((count (fix:- end start))
	   (async-buffer (ensure-buffer gstream count)))
      (set-gio-cleanup-info-pending-op! gio-info 'write)
      (c-poke-bytes async-buffer 0 count buffer start)
      (C-call "g_output_stream_write_async"
	      (gobject-alien gstream)
	      async-buffer
	      count
	      (gio-priority gstream)
	      (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	      (C-callback "async_ready")
	      callback-id)
      (let* ((queue (gio-queue gstream))
	     (value (dequeue! queue)))
	(if (string? value)
	    (begin
	      (set-gio-cleanup-info-pending-op! gio-info 'error)
	      (error "Error writing:" gstream value))
	    (begin
	      (set-gio-cleanup-info-pending-op! gio-info #f)
	      value))))))

(define (make-write-finish-callback queue gerror*)
  (C-callback
   (named-lambda (write-finish-callback source result)
     (assert-glib-locked 'write-finish-callback)
     (let ((bytes
	    (C-call "g_output_stream_write_finish" source result gerror*)))
       (if (fix:= bytes -1)
	   (let ((message (%gerror-message gerror*)))
	     (%trace ";write-finish-callback "message" "queue"\n")
	     (%queue! queue message))
	   (begin
	     (%trace ";write-finish-callback "bytes" "queue"\n")
	     (%queue! queue bytes)))))))

(define (g-output-stream-flush gstream)
  (assert-glib-locked 'g-output-stream-flush)
  (let* ((gio-info (gio-cleanup-info gstream))
	 (info (g-output-stream-cleanup-info gstream))
	 (callback-id (g-output-stream-cleanup-info-flush-id info)))
    (guarantee-gio-idle gstream)
    (set-gio-cleanup-info-pending-op! gio-info 'flush)
    (C-call "g_output_stream_flush_async"
	    (gobject-alien gstream)
	    (gio-priority gstream)
	    (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	    (C-callback "async_ready")
	    callback-id)
    (let* ((queue (gio-queue gstream))
	   (value (dequeue! queue)))
      (if (string? value)
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info 'error)
	    (error "Error flushing:" gstream value))
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info #f)
	    (not (fix:zero? value)))))))

(define (make-flush-finish-callback queue gerror*)
  (C-callback
   (named-lambda (flush-finish-callback source result)
     (assert-glib-locked 'flush-finish-callback)
     (if (fix:zero?
	  (C-call "g_output_stream_flush_finish" source result gerror*))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";flush-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";flush-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define (g-output-stream-close gstream)
  (assert-glib-locked 'g-output-stream-close)
  (gfile-close gstream
	       (named-lambda (close-output
			      gstream* priority gcancellable* callback id)
		 (C-call "g_output_stream_close_async"
			 gstream* priority gcancellable* callback id))
	       make-output-close-finish-callback
	       (named-lambda (close-output-cleanup gio-info)
		 (free (g-stream-buffer gstream))
		 (cleanup-g-output-stream
		  gio-info (g-output-stream-cleanup-info gstream)))))

(define (make-output-close-finish-callback queue gerror*)
  (C-callback
   (named-lambda (output-close-finish-callback source result)
     (assert-glib-locked 'output-close-finish-callback)
     (if (fix:zero?
	  (C-call "g_output_stream_close_finish" source result gerror*))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";output-close-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";output-close-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define-class <gfile-input-stream>
    (<g-input-stream>))

(define-method initialize-instance ((gstream <gfile-input-stream>))
  (call-next-method gstream)
  (let ((alien (gobject-alien gstream)))
    (set-alien/ctype! alien '|GFileInputStream|)))

(define (gfile-read gfile)
  (assert-glib-locked 'gfile-read)
  (gfile-open gfile 'open
	      make-g-input-stream
	      (named-lambda (open-callout
			     gfile* priority gcancellable* callback id)
		(C-call "g_file_read_async"
			gfile* priority gcancellable* callback id))
	      make-open-finish-callback
	      setup-input))

(define (gfile-open gfile operation make-gstream callout make-callback setup)
  (let* ((gstream (make-gstream))
	 (gio-info (gio-cleanup-info gstream))
	 (queue (gio-queue gstream))
	 (gerror* (gio-cleanup-info-gerror-pointer gio-info))
	 (callback-id
	  (without-interruption		;don't leak callback IDs
	   (lambda ()
	     (let ((id (make-callback (gobject-alien gstream) queue gerror*)))
	       (set-gio-cleanup-info-pending-op! gio-info operation)
	       (set-gio-cleanup-info-callback-id! gio-info id)
	       id)))))
    (let retry ()
      (callout (gobject-alien gfile)
	       (gio-priority gstream)
	       (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	       (C-callback "async_ready")
	       callback-id)
      (let ((value (dequeue! queue)))
	(cond ((eq? value #t)
	       (set-gio-cleanup-info-pending-op! gio-info #f)
	       (without-interruption
		(lambda ()
		  (de-register-c-callback callback-id)
		  (set-gio-cleanup-info-callback-id! gio-info #f)
		  (setup gstream queue gerror*)))
	       gstream)
	      ((equal? value "The specified location is not mounted")
	       (gfile-mount gfile)
	       (retry))
	      ((string? value)
	       (set-gio-cleanup-info-pending-op! gio-info 'error)
	       (error (string-append (uri->string (gfile-uri gfile))":") value))
	      (else
	       (error "Unexpected:" value gstream)))))))

(define (make-open-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (open-finish-callback source result)
     (assert-glib-locked 'open-finish-callback)
     (C-call "g_file_read_finish" alien source result gerror*)
     (if (alien-null? alien)
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";open-finish-callback \""message"\" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";open-finish-callback "alien" "queue"\n")
	   (%queue! queue #t))))))

(define (setup-input gstream queue gerror*)
  (let ((info (g-input-stream-cleanup-info gstream)))
    (set-g-input-stream-cleanup-info-read-id!
     info (make-read-finish-callback queue gerror*))
    (set-g-input-stream-cleanup-info-skip-id!
     info (make-skip-finish-callback queue gerror*))))

(define-class <gfile-output-stream>
    (<g-output-stream>))

(define-method initialize-instance ((gstream <gfile-output-stream>))
  (call-next-method gstream)
  (let ((alien (gobject-alien gstream)))
    (set-alien/ctype! alien '|GFileOutputStream|)))

(define (gfile-append-to gfile . flags)
  (assert-glib-locked 'gfile-append-to)
  (let ((flags* (->gfile-create-flags flags)))
    (gfile-open gfile 'append-to
		make-g-output-stream
		(named-lambda (append-to-callout
			       gfile* priority gcancellable* callback id)
		  (C-call "g_file_append_to_async"
			  gfile* flags* priority gcancellable* callback id))
		make-append-to-finish-callback
		setup-output)))

(define (->gfile-create-flags flags)
  (apply bit-ior (map! ->gfile-create-flag flags)))

(define (->gfile-create-flag flag)
  (case flag
    ((private) (C-enum "G_FILE_CREATE_PRIVATE"))
    ((replace) (C-enum "G_FILE_CREATE_REPLACE_DESTINATION"))
    (else (error:wrong-type-argument flag "GFile create flag"
				     '->gfile-create-flag))))

(define (make-append-to-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (append-to-finish-callback source result)
     (assert-glib-locked 'append-to-finish-callback)
     (C-call "g_file_append_to_finish" alien source result gerror*)
     (g-output-stream-finish alien queue gerror* 'append-to))))

(define (g-output-stream-finish alien queue gerror* op)
  (assert-glib-locked 'g-output-stream-finish)
  (if (alien-null? alien)
      (let ((message (%gerror-message gerror*)))
	(%trace ";"op"-finish-callback "message" "queue"\n")
	(%queue! queue message))
      (begin
	(%trace ";"op"-finish-callback "alien" "queue"\n")
	(%queue! queue #t))))

(define (setup-output gstream queue gerror*)
  (let ((info (g-output-stream-cleanup-info gstream)))
    (set-g-output-stream-cleanup-info-write-id!
     info (make-write-finish-callback queue gerror*))
    (set-g-output-stream-cleanup-info-flush-id!
     info (make-flush-finish-callback queue gerror*))))

(define (gfile-create gfile . flags)
  (assert-glib-locked 'gfile-create)
  (let ((flags* (->gfile-create-flags flags)))
    (gfile-open gfile 'create
		make-g-output-stream
		(named-lambda (create-callout
			       gfile* priority gcancellable* callback id)
		  (C-call "g_file_create_async"
			  gfile* flags* priority gcancellable* callback id))
		make-create-finish-callback
		setup-output)))

(define (make-create-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (create-finish-callback source result)
     (assert-glib-locked 'create-finish-callback)
     (C-call "g_file_create_finish" alien source result gerror*)
     (g-output-stream-finish alien queue gerror* 'create))))

(define (gfile-replace gfile etag backup? . flags)
  (assert-glib-locked 'gfile-replace)
  (let ((etag (->gfile-etag etag))
	(make-backups (if backup? 1 0))
	(flags* (->gfile-create-flags flags)))
    (gfile-open gfile 'replace
		make-g-output-stream
		(named-lambda (replace-callout
			       gfile* priority gcancellable* callback id)
		  (C-call "g_file_replace_async"
			  gfile* etag make-backups flags*
			  priority gcancellable* callback id))
		make-replace-finish-callback
		setup-output)))

(define-integrable (->gfile-etag etag)
  (cond ((and (alien? etag) (eq? (alien/ctype etag) '|GFile etag|))
	 etag)
	((or (eq? etag #f) (zero? etag))
	 0)
	(else
	 (error:wrong-type-argument etag "GFile etag" '->gfile-etag))))

(define (make-replace-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (replace-finish-callback source result)
     (assert-glib-locked 'replace-finish-callback)
     (C-call "g_file_replace_finish" alien source result gerror*)
     (g-output-stream-finish alien queue gerror* 'replace))))

(define-class (<gfile-info> (constructor ()))
    (<gio>))

(define-method initialize-instance ((object <gfile-info>))
  (call-next-method object)
  (add-glib-cleanup object (make-ginfo-cleanup (gio-cleanup-info object))))

(define (make-ginfo-cleanup gio-info)
  (named-lambda (ginfo-cleanup)
    (cleanup-gio gio-info)))

(define (gfile-query-info gfile attributes follow-symlinks?)
  (guarantee string? attributes 'gfile-query-info)
  (assert-glib-locked 'gfile-query-info)
  (gfile-open gfile 'query
	      make-gfile-info
	      (named-lambda (query-callout
			     gfile* priority gcancellable* callback id)
		(C-call "g_file_query_info_async"
			gfile*
			attributes
			(if follow-symlinks?
			    (C-enum "G_FILE_QUERY_INFO_NONE")
			    (C-enum "G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS"))
			priority gcancellable* callback id))
	      make-query-finish-callback 
	      (named-lambda (cleanup-gfile-info gfile-info queue gerror*)
		(declare (ignore queue gerror*))
		(let ((gio-info (gio-cleanup-info gfile-info)))
		  (set-gio-cleanup-info-pending-op! gio-info 'closed)
		  (cleanup-gio gio-info)))))

(define (make-query-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (query-finish-callback source result)
     (assert-glib-locked 'query-finish-callback)
     (C-call "g_file_query_info_finish" alien source result gerror*)
     (if (alien-null? alien)
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";query-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";query-finish-callback "alien" "queue"\n")
	   (%queue! queue #t))))))

(define (gfile-info-list-attributes ginfo namespace)
  (guarantee string? namespace 'gfile-info-list-attributes)
  (assert-glib-locked 'gfile-info-list-attributes)
  (map! string->symbol
	(let ((alien (make-cstringv
		      (lambda (copy)
			(C-call "g_file_info_list_attributes" copy
				(gobject-alien ginfo) namespace)))))
	  (let ((strings (peek-cstringv alien)))
	    (free-cstringv alien)
	    strings))))

(define (gfile-info-get-attribute-status ginfo name)
  (assert-glib-locked 'gfile-info-get-attribute-status)
  (let ((code (C-call "g_file_info_get_attribute_status"
		      (gobject-alien ginfo)
		      (string->utf8 name))))
    (cond ((fix:= code (C-enum "G_FILE_ATTRIBUTE_STATUS_UNSET")) 'unset)
	  ((fix:= code (C-enum "G_FILE_ATTRIBUTE_STATUS_SET")) 'set)
	  ((fix:= code (C-enum "G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING"))
	   'error-setting)
	  (else (error "Unknown GFileAttributeStatus:" code)))))

(define (gfile-info-get-attribute-value ginfo name)
  (assert-glib-locked 'gfile-info-get-attribute-value)
  (let* ((alien (gobject-alien ginfo))
	 (name-bv (string->utf8 name))
	 (type (C-call "g_file_info_get_attribute_type" alien name-bv)))
    (cond ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_INVALID"))
	   (error "Invalid attribute:" name))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_STRING"))
	   (c-peek-cstring
	    (C-call "g_file_info_get_attribute_string"
		    (make-alien 'char) alien name-bv)))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_BYTE_STRING"))
	   (c-peek-cstring
	    (C-call "g_file_info_get_attribute_byte_string"
		    (make-alien 'uchar) alien name-bv)))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_BOOLEAN"))
	   (not (fix:zero?
		 (C-call "g_file_info_get_attribute_boolean" alien name-bv))))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_UINT32"))
	   (C-call "g_file_info_get_attribute_uint32" alien name-bv))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_INT32"))
	   (C-call "g_file_info_get_attribute_int32" alien name-bv))
;	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_UINT64"))
;	   (C-call "g_file_info_get_attribute_uint64" alien name-bv))
;	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_INT64"))
;	   (C-call "g_file_info_get_attribute_int64" alien name-bv))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_OBJECT"))
	   (C-call "g_file_info_get_attribute_object"
		   (make-alien '|GObject|) alien name-bv))
	  ((fix:= type (C-enum "G_FILE_ATTRIBUTE_TYPE_STRINGV"))
	   (peek-cstringv
	    (C-call "g_file_info_get_attribute_stringv"
		    (make-alien '(* (const char))) alien name-bv)))
	  (else (error "Unexpected GFileAttributeType:" type)))))

(define-class (<gfile-enumerator> (constructor ()))
    (<gio>)
  ;; Nascent gfile-enumerator-cleanup.  Just a GList at the mo'.
  (ginfos
   define accessor initializer (lambda () (make-alien '|GList|))))

(define-method initialize-instance ((object <gfile-enumerator>))
  (call-next-method object)
  (add-glib-cleanup object (make-gfile-enumerator-cleanup
			    (gio-cleanup-info object)
			    (gfile-enumerator-ginfos object))))

(define (make-gfile-enumerator-cleanup gio-info ginfos)
  (named-lambda (gfile-enumerator-cleanup)
    (cleanup-gfile-enumerator gio-info ginfos)))

(define (cleanup-gfile-enumerator gio-info ginfos)
  (assert-without-interruption 'cleanup-gfile-enumerator)
  (assert-glib-locked 'cleanup-gfile-enumerator)
  (cleanup-gio gio-info)
  (cleanup-ginfos ginfos))

(define (cleanup-ginfos glist)
  (assert-without-interruption 'cleanup-ginfos)
  (assert-glib-locked 'cleanup-ginfos)
  (if (not (alien-null? glist))
      (let ((scan (copy-alien glist))
	    (ginfo (make-alien '|GFileInfo|)))
	(let loop ()
	  (C-> scan "GList data" ginfo)
	  (if (not (alien-null? ginfo))
	      (begin
		(C->= scan "GList data" 0)
		(C-call "g_object_unref" ginfo)))
	  (C-> scan "GList next" scan)
	  (if (alien-null? scan)
	      (begin
		(C-call "g_list_free" glist)
		(alien-null! glist))
	      (loop))))))

(define (gfile-enumerate-children gfile attributes follow-symlinks?)
  (guarantee string? attributes 'gfile-enumerate-children)
  (assert-glib-locked 'gfile-enumerate-children)
  (gfile-open gfile 'open
	      make-gfile-enumerator
	      (named-lambda (query-callout
			     gfile* priority gcancellable* callback id)
		(C-call "g_file_enumerate_children_async"
			gfile*
			(string->utf8 attributes)
			(if follow-symlinks?
			    (C-enum "G_FILE_QUERY_INFO_NONE")
			    (C-enum "G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS"))
			priority gcancellable* callback id))
	      make-enumerator-finish-callback
	      setup-enumerator))

(define (setup-enumerator genum queue gerror*)
  (let ((info (gio-cleanup-info genum)))
    (set-gio-cleanup-info-callback-id!
     info (make-next-files-finish-callback
	   (gfile-enumerator-ginfos genum) queue gerror*))))

(define (make-enumerator-finish-callback alien queue gerror*)
  (C-callback
   (named-lambda (enumerator-finish-callback source result)
     (assert-glib-locked 'enumerator-finish-callback)
     (C-call "g_file_enumerate_children_finish" alien source result gerror*)
     (if (alien-null? alien)
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";enumerator-finish-callback \""message"\" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";enumerator-finish-callback "alien" "queue"\n")
	   (%queue! queue #t))))))

(define (gfile-enumerator-next-files genum nfiles)
  (guarantee fixnum? nfiles 'gfile-enumerator-next-files)
  (assert-glib-locked 'gfile-enumerator-next-files)
  (let* ((gio-info (gio-cleanup-info genum))
	 (callback-id (gio-cleanup-info-callback-id gio-info)))
    (guarantee-gio-idle genum)
    (set-gio-cleanup-info-pending-op! gio-info 'next)
    (C-call "g_file_enumerator_next_files_async"
	    (gobject-alien genum)
	    nfiles
	    (gio-priority genum)
	    (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	    (C-callback "async_ready")
	    callback-id)
    (let* ((queue (gio-queue genum))
	   (value (dequeue! queue)))
      (if (string? value)
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info 'error)
	    (error "Error in gfile-enumerator-next-files:" genum value))
	  (begin
	    (set-gio-cleanup-info-pending-op! gio-info #f)
	    (make-ginfos genum))))))

(define (make-ginfos genum)
  (assert-glib-locked 'make-ginfos)
  (let* ((glist (gfile-enumerator-ginfos genum))
	 (scan (copy-alien glist))
	 (ginfo (make-alien '|GFileInfo|))
	 (ginfos
	  (let loop ()
	    (if (alien-null? scan)
		'()
		(begin
		  (C-> scan "GList data" ginfo)
		  (if (not (alien-null? ginfo))
		      (let ((new (make-gfile-info)))
			(copy-alien-address! (gobject-alien new) ginfo)
			(C->= scan "GList data" 0)
			(C-> scan "GList next" scan)
			(cons new (loop)))
		      (begin
			(C-> scan "GList next" scan)
			(loop))))))))
    (if (not (alien-null? glist))
	(begin
	  (C-call "g_list_free" glist)
	  (alien-null! glist)))
    ginfos))

(define (make-next-files-finish-callback ginfos queue gerror*)
  (C-callback
   (named-lambda (next-files-finish-callback source result)
     (assert-glib-locked 'next-files-finish-callback)
     (C-call "g_file_enumerator_next_files_finish" ginfos source result gerror*)
     (if (and (alien-null? ginfos)
	      (not (alien-null? (C-> gerror* "* GError"))))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";next-files-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";next-files-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define (gfile-enumerator-close genum)
  (assert-glib-locked 'gfile-enumerator-close)
  (let ((ginfos (gfile-enumerator-ginfos genum)))
    (gfile-close genum
		 (named-lambda (close-enumerator
				genum* priority gcancellable* callback id)
		   (C-call "g_file_enumerator_close_async"
			   genum* priority gcancellable* callback id))
		 make-enumerator-close-finish-callback
		 (named-lambda (cleanup-enumerator gio-info)
		   (cleanup-gfile-enumerator gio-info ginfos)))))

(define (make-enumerator-close-finish-callback queue gerror*)
  (C-callback
   (named-lambda (enumerator-close-finish-callback source result)
     (assert-glib-locked 'enumerator-close-finish-callback)
     (if (fix:zero?
	  (C-call "g_file_enumerator_close_finish" source result gerror*))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";enumerator-close-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";enumerator-close-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define-class (<g-mount-operation> (constructor ()))
    (<gio>)
  (message define standard initial-value #f)
  (username define standard initial-value #f)
  (domain define standard initial-value #f)
  (ask-password-flags define standard initial-value #f))

(define-method initialize-instance ((gmountop <g-mount-operation>))
  (call-next-method gmountop)
  (add-glib-cleanup gmountop
		    (make-gmountop-cleanup (gio-cleanup-info gmountop)))
  (let ((alien (gobject-alien gmountop)))
    (set-alien/ctype! alien '|GMountOperation|)
    (C-call "g_mount_operation_new" alien)
    (error-if-null alien "Could not create:" gmountop)))

(define (make-gmountop-cleanup gio-info)
  (named-lambda (gmountop-cleanup)
    (cleanup-gio gio-info)))

(define (gfile-mount gfile)
  (assert-glib-locked 'gfile-mount)
  (let* ((gmountop (make-g-mount-operation))
	 (alien (gobject-alien gmountop))
	 (gio-info (gio-cleanup-info gmountop))
	 (queue (gio-queue gmountop))
	 (gerror* (gio-cleanup-info-gerror-pointer gio-info))
	 (callback-id
	  (without-interruption		;don't leak callback IDs
	   (lambda ()
	     (let ((id (make-mount-finish-callback queue gerror*)))
	       (set-gio-cleanup-info-pending-op! gio-info 'mount)
	       (set-gio-cleanup-info-callback-id! gio-info id)
	       id)))))
    (let ((userinfo (uri-authority-userinfo (uri-authority (gfile-uri gfile)))))
      (if userinfo
	  (if (string=? userinfo "anonymous")
	      (begin
		(C-call "g_mount_operation_set_anonymous" alien 1)
		(set-g-mount-operation-username! gmountop "anonymous"))
	      (begin
		(C-call "g_mount_operation_set_username"
			alien (string->utf8 userinfo))
		(set-g-mount-operation-username! gmountop userinfo)))))
    (attach-mount-signal-handlers gmountop gfile)
    (let retry ()
      (C-call "g_file_mount_enclosing_volume"
	      (gobject-alien gfile)
	      (C-enum "G_MOUNT_MOUNT_NONE")
	      alien
	      (gobject-alien (gio-cleanup-info-gcancellable gio-info))
	      (C-callback "async_ready")
	      callback-id)
      (let ((value (dequeue! queue)))
	(cond ((and (equal? value "Password dialog cancelled")
		    (g-mount-operation-ask-password-flags gmountop))
	       (set-gio-cleanup-info-pending-op! gio-info #f)
	       (prompt-for-mount-auth gmountop)
	       (retry))
	      ((string? value)
	       (set-gio-cleanup-info-pending-op! gio-info 'error)
	       (error (string-append (uri->string (gfile-uri gfile))":") value))
	      ((eq? value #t)
	       (set-gio-cleanup-info-pending-op! gio-info #f)
	       (gobject-unref! gmountop)
	       (without-interruption
		(lambda ()
		  (cleanup-gio gio-info)))
	       unspecific)
	      (else
	       (error "Unexpected value:" value gmountop)))))))

(define (prompt-for-mount-auth gmountop)
  (%trace-auth ";prompt-for-mount-auth "gmountop"\n")
  (assert-glib-locked 'prompt-for-mount-auth)
  (let ((message (g-mount-operation-message gmountop))
	(domain (g-mount-operation-domain gmountop))
	(username (g-mount-operation-username gmountop))
	(flags (g-mount-operation-ask-password-flags gmountop))
	(alien (gobject-alien gmountop))
	(port (interaction-i/o-port)))
    (if message (display message port))
    (if (and (memq 'anonymous-supported flags)
	     (without-glib-lock
	       (lambda ()
		 (prompt-for-confirmation "Login anonymously" port))))
	(begin
	  (C-call "g_mount_operation_set_anonymous" alien 1)
	  (set-g-mount-operation-username! alien "anonymous")))
    (if (memq 'need-domain flags)
	(let ((d (prompt-for-string* "Domain" domain port)))
	  (C-call "g_mount_operation_set_domain" alien (string->utf8 d))
	  (set-g-mount-operation-domain! gmountop d)))
    (if (memq 'need-username flags)
	(let ((u (prompt-for-string* "Username" username port)))
	  (C-call "g_mount_operation_set_username" alien (string->utf8 u))
	  (set-g-mount-operation-username! gmountop u)))
    (if (memq 'need-password flags)
	(let ((password))
	  (dynamic-wind
	   (lambda () unspecific)
	   (lambda ()
	     (without-glib-lock
	      (lambda ()
		(call-with-pass-phrase
		 "Password"
		 (lambda (phrase) (set! password (string->utf8 phrase))))))
	     (C-call "g_mount_operation_set_password" alien password))
	   (lambda ()
	     (bytevector-fill! password #x55)))))
    (if (memq 'saving-supported flags)
	(if (without-glib-lock
	     (lambda ()
	       (prompt-for-confirmation "Save password permanently" port)))
	    (C-call "g_mount_operation_set_password_save" alien
		    (C-enum "G_PASSWORD_SAVE_PERMANENTLY"))
	    (if (without-glib-lock
		 (lambda ()
		   (prompt-for-confirmation "Save password for this session"
					    port)))
		(C-call "g_mount_operation_set_password_save" alien
			(C-enum "G_PASSWORD_SAVE_FOR_SESSION"))
		(C-call "g_mount_operation_set_password_save" alien
			(C-enum "G_PASSWORD_SAVE_NEVER")))))))

(define (prompt-for-string* prompt default port)
  (let ((s (without-glib-lock
	    (lambda ()
	      (prompt-for-string prompt port)))))
    (if (not (string-find-next-char-in-set s char-set:not-whitespace))
	default
	s)))

(define (make-mount-finish-callback queue gerror*)
  (C-callback
   (named-lambda (mount-finish-callback source result)
     (assert-glib-locked 'mount-finish-callback)
     (if (fix:zero? (C-call "g_file_mount_enclosing_volume_finish"
			    source result gerror*))
	 (let ((message (%gerror-message gerror*)))
	   (%trace ";mount-finish-callback "message" "queue"\n")
	   (%queue! queue message))
	 (begin
	   (%trace ";mount-finish-callback #t "queue"\n")
	   (%queue! queue #t))))))

(define (attach-mount-signal-handlers gmountop gfile)
  (g-signal-connect gmountop (C-callback "ask_password")
		    (make-mount-password-callback gfile))
  (g-signal-connect gmountop (C-callback "ask_question")
		    (make-mount-question-callback gfile))
  (g-signal-connect gmountop (C-callback "show_processes")
		    (make-mount-processes-callback gfile)))

(define (make-mount-password-callback gfile)
  (named-lambda (mount-password-callback gmountop message user domain flags)
    (%trace-auth ";mount-password-callback "(gfile-uri gfile)
		 " "gmountop
		 " "(c-peek-cstring message)
		 " "(c-peek-cstring user)
		 " "(c-peek-cstring domain)
		 " "(->ask-password-flags flags)"\n")
    (assert-glib-locked 'mount-password-callback)
    (let ((old (g-mount-operation-ask-password-flags gmountop))
	  (new (->ask-password-flags flags)))
      (set-g-mount-operation-message! gmountop
				      (c-peek-cstring message))
      (set-g-mount-operation-username! gmountop
				       (c-peek-cstring user))
      (set-g-mount-operation-domain! gmountop
				     (c-peek-cstring domain))
      (set-g-mount-operation-ask-password-flags! gmountop new)
      (cond ((not old)
	     ;; Punt, %queuing "Password dialog cancelled".
	     (C-call "g_mount_operation_reply" (gobject-alien gmountop)
		     (C-enum "G_MOUNT_OPERATION_UNHANDLED")))
	    ((equal? old new)
	     ;; Assume gmountop is now setup after "...dialog cancelled".
	     (C-call "g_mount_operation_reply" (gobject-alien gmountop)
		     (C-enum "G_MOUNT_OPERATION_HANDLED")))))))

(define (->ask-password-flags flags)
  (define-integrable (cons-flags mask symbol rest)
    (if (bit? flags mask)
	(cons symbol rest)
	rest))
  (cons-flags
   (C-enum "G_ASK_PASSWORD_NEED_PASSWORD") 'need-password
   (cons-flags
    (C-enum "G_ASK_PASSWORD_NEED_USERNAME") 'need-username
    (cons-flags
     (C-enum "G_ASK_PASSWORD_NEED_DOMAIN") 'need-domain
     (cons-flags
      (C-enum "G_ASK_PASSWORD_SAVING_SUPPORTED") 'saving-supported
      (cons-flags
       (C-enum "G_ASK_PASSWORD_ANONYMOUS_SUPPORTED") 'anonymous-supported
       '()))))))

(define (make-mount-question-callback gfile)
  (named-lambda (mount-question-callback gmountop message choices)
    (%trace-auth ";make-mount-question-callback"
		 " "(gfile-uri gfile)
		 " "gmountop
		 " "(c-peek-cstring message)
		 " "(peek-gstrv! choices)"\n")
    (warn "Unimplemented" 'mount-question-callback)
    (assert-glib-locked 'mount-question-callback)))

(define (peek-gstrv! alien)
  (let loop ()
    (let ((str (c-peek-cstringp! alien)))
      (if (null? str)
	  '()
	  (cons str (loop))))))

(define (make-mount-processes-callback gfile)
  (named-lambda (mount-processes-callback gmountop message processes choices)
    (%trace-auth ";make-mount-processes-callback"
		 " "gfile
		 " "gmountop
		 " "(c-peek-cstring message)
		 " "processes
		 " "(peek-gstrv! choices)"\n")
    (warn "Unimplemented" 'mount-processes-callback)))

(define-class (<gfile> (constructor (uri)))
    (<gobject>)
  (uri define accessor))

(define-method initialize-instance ((gfile <gfile>))
  (call-next-method gfile)
  (assert-glib-locked '(initialize-instance <gfile>))
  (let ((alien (gobject-alien gfile))
	(uri (string->utf8 (uri->string (gfile-uri gfile)))))
    (set-alien/ctype! alien '|GFile|)
    (C-call "g_file_new_for_uri" alien uri)
    (error-if-null alien "Could not create:" gfile uri)))

(define-class (<gcancellable> (constructor ()))
    (<gobject>))

(define-method initialize-instance ((gcancel <gcancellable>))
  (call-next-method gcancel)
  (assert-glib-locked '(initialize-instance <gcancellable>))
  (let ((alien (gobject-alien gcancel)))
    (set-alien/ctype! alien '|GCancellable|)
    (C-call "g_cancellable_new" alien)))

(define-structure gfile-etag
  alien)

(define (make-cstringv setter)
  ;; SETTER is applied to an alien that must not escape.
  (assert-glib-locked 'make-cstringv)
  (let ((alien (make-alien '(* uchar)))
	(copy (make-alien '(* uchar))))
    (add-glib-cleanup alien (make-cstringv-cleanup copy))
    (setter copy)
    (copy-alien-address! alien copy)
    alien))

(define (make-cstringv-cleanup alien)
  (named-lambda (cstringv-cleanup)
    (assert-glib-locked 'cstringv-cleanup)
    (if (not (alien-null? alien))
	(let ((scan (copy-alien alien))
	      (cstr (make-alien 'uchar)))
	  (let loop ()
	    (C-> scan "* uchar" cstr)
	    (if (not (alien-null? cstr))
		(begin
		  (C-call "g_free" cstr)
		  (alien-byte-increment! scan (C-sizeof "* uchar"))
		  (loop))))
	  (C-call "g_free" alien)
	  (alien-null! alien)))))

(define (peek-cstringv alien)
  (let ((scan (copy-alien alien))
	(cstr (make-alien 'uchar)))
    (let loop ()
      (C-> scan "* uchar" cstr)
      (if (alien-null? cstr)
	  '()
	  (let ((str (c-peek-cstring cstr)))
	    (alien-byte-increment! scan (C-sizeof "* uchar"))
	    (cons str (loop)))))))

(define (free-cstringv alien)
  (without-interruption
   (lambda ()
     (execute-glib-cleanup alien)
     (alien-null! alien))))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ args ...)
     (if %trace? (outf-error args ...)))))

(define %trace-auth? #t)

(define-syntax %trace-auth
  (syntax-rules ()
    ((_ args ...)
     (if %trace? (outf-error args ...)))))