#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a GLib plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; GObjects
;;; package: (glib gobject)

(C-include "glib")

(define-class <gobject> ()

  ;; The address of the toolkit object.  A null alien if the GObject
  ;; has not been created (yet), or has been unrefed.
  (alien define accessor
	 initializer (lambda () (make-alien '|GObject|)))

  ;; A pair, shared with cleanup thunks.  The cdr of this
  ;; pair is the alist associating signal names with Scheme callback
  ;; IDs and toolkit handles.  In this alist, a callback ID will be #f
  ;; if the signal was disconnected.
  (signals define standard
	   initializer (lambda () (list 'GOBJECT-SIGNALS)))

  ;; This instance's weak-pair on the glib-cleanups list.  This is
  ;; cached here mainly for g-signal-connect, which must create
  ;; callbacks that only weakly reference this instance.
  (weak-self define standard))

(define-guarantee gobject "a <gobject>")

(define-integrable (gobject-live? object)
  (not (alien-null? (gobject-alien object))))

(define-method initialize-instance ((object <gobject>))
  (call-next-method object)
  (set-gobject-weak-self!
   object (add-glib-cleanup object (make-gobject-cleanup
				    (gobject-alien object)
				    (gobject-signals object)))))

(define (make-gobject-cleanup alien signals)
  ;; This separate procedure ensures that the gobject is not caught in
  ;; the closure.
  (named-lambda (gobject-cleanup)
    (%trace ";gobject-cleanup "alien"\n")
    (assert-glib-locked 'gobject-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (for-each
	    (lambda (name.id.handle) (disconnect!? alien (cdr name.id.handle)))
	    (cdr signals))
	  (C-call "g_object_unref" alien)
	  (alien-null! alien)))
    (%trace ";gobject-cleanup done with "alien"\n")))

(define (gobject-unref! object)
  (assert-glib-locked 'gobject-unref!)
  (without-interruption
   (lambda ()
     (execute-glib-cleanup object))))

(define (g-signal-connect gobject alien-function callback
			  #!optional signal-name)
  ;; Specify signal-name if it is not the same as alien-function's name.
  (guarantee-gobject gobject 'g-signal-connect)
  (guarantee-alien-function alien-function 'g-signal-connect)
  (assert-glib-locked 'g-signal-connect)
  (let ((name (cond ((default-object? signal-name)
		     (string->symbol (alien-function/name alien-function)))
		    ((symbol? signal-name) signal-name)
		    ((string? signal-name) (string->symbol signal-name))
		    (else
		     (error:wrong-type-argument
		      signal-name "string or symbol" 'g-signal-connect)))))
    (without-interruption
     (lambda ()
       (let* ((alien (gobject-alien gobject))
	      (signals (gobject-signals gobject))
	      (name.id.handle
	       (or (assq name (cdr signals))
		   (let ((entry (cons* name #f #f)))
		     (set-cdr! signals (cons entry (cdr signals)))
		     entry))))
	 (disconnect!? alien (cdr name.id.handle))
	 (connect! alien name.id.handle
		   alien-function
		   (register-c-callback
		    (make-gobject-signal-callback
		     name (gobject-weak-self gobject) callback))))))))

(define (make-gobject-signal-callback name weak-pair callback)
  (named-lambda (gobject-signal-callback instance . args)
    ;; Callbacks run without-interrupts.
    (if (weak-pair/car? weak-pair)
	(let ((gobject (weak-car weak-pair)))
	  (if (not (alien=? (gobject-alien gobject) instance))
	      (warn "Signal instance / gobject mismatch:" instance gobject))
	  (apply callback gobject args))
	(error "Cannot signal a <gobject> that is already GC'ed:" name args))))

(define (connect! alien name.id.handle alien-function newid)
  (let ((id.handle (cdr name.id.handle)))
    (set-car! id.handle newid)
    (set-cdr! id.handle
	      (C-call "g_signal_connect_data" alien
		      (string->utf8 (symbol->string (car name.id.handle)))
		      alien-function newid 0 0))))

(define (g-signal-disconnect gobject name)
  (guarantee-gobject gobject 'g-signal-disconnect)
  (guarantee symbol? name 'g-signal-disconnect)
  (assert-glib-locked 'g-signal-disconnect)
  (without-interruption
   (lambda ()
     (let* ((alien (gobject-alien gobject))
	    (signals (gobject-signals gobject))
	    (name.id.handle (assq name (cdr signals))))
	 (if (not name.id.handle)
	     (warn "No signal to disconnect:" name gobject)
	     (if (not (disconnect!? alien (cdr name.id.handle)))
		 (warn "Signal already disconnected:" name gobject)))))))

(define (disconnect!? alien id.handle)
  (if (eq? (car id.handle) #f)
      #f
      (begin
	(C-call "g_signal_handler_disconnect" alien (cdr id.handle))
	(de-register-c-callback (car id.handle))
	(set-car! id.handle #f)
	#t)))


;;; Properties

(define (gobject-get-property gobject property)
  (guarantee-gobject gobject 'gobject-get-property)
  (assert-glib-locked 'gobject-get-property)

  (let ((name (check-prop-name property))
	(gvalue (malloc (C-sizeof "GValue") '|GValue|)))

    (define (unimplemented type)
      (error "Unimplemented property type:" type name gobject))

    (C-call "g_object_get_property" (gobject-alien gobject)
	    (string->utf8 name) gvalue)
    (let* ((type (C-> gvalue "GValue g_type"))
	   (value
	    (cond
	     ((int:= type (C-enum "G_TYPE_INVALID"))
	      (error "Invalid property:" name gobject))
	     ((int:= type (C-enum "G_TYPE_NONE"))
	      (error "Void property:" name gobject))
	     ((int:= type (C-enum "G_TYPE_INTERFACE"))
	      (unimplemented "an interface"))
	     ((int:= type (C-enum "G_TYPE_CHAR"))
	      (C-call "g_value_get_schar" gvalue))
	     ((int:= type (C-enum "G_TYPE_UCHAR"))
	      (C-call "g_value_get_uchar" gvalue))
	     ((int:= type (C-enum "G_TYPE_BOOLEAN"))
	      (C-call "g_value_get_boolean" gvalue))
	     ((int:= type (C-enum "G_TYPE_INT"))
	      (C-call "g_value_get_int" gvalue))
	     ((int:= type (C-enum "G_TYPE_UINT"))
	      (C-call "g_value_get_uint" gvalue))
	     ((int:= type (C-enum "G_TYPE_LONG"))
	      (C-call "g_value_get_long" gvalue))
	     ((int:= type (C-enum "G_TYPE_ULONG"))
	      (C-call "g_value_get_ulong" gvalue))
;	     ((int:= type (C-enum "G_TYPE_INT64"))
;	      (C-call "g_value_get_int64" gvalue))
;	     ((int:= type (C-enum "G_TYPE_UINT64"))
;	      (C-call "g_value_get_uint64" gvalue))
	     ((int:= type (C-enum "G_TYPE_ENUM"))
	      (C-call "g_value_get_enum" gvalue))
	     ((int:= type (C-enum "G_TYPE_FLAGS"))
	      (C-call "g_value_get_flags" gvalue))
	     ((int:= type (C-enum "G_TYPE_FLOAT"))
	      (C-call "g_value_get_float" gvalue))
	     ((int:= type (C-enum "G_TYPE_DOUBLE"))
	      (C-call "g_value_get_double" gvalue))
	     ((int:= type (C-enum "G_TYPE_STRING"))
	      (let ((alien (make-alien '(const (* |gchar|)))))
		(C-call "g_value_get_string" alien gvalue)
		(let ((str (c-peek-cstring alien)))
		  (free alien)
		  str)))
	     ((int:= type (C-enum "G_TYPE_POINTER"))
	      (let ((alien (make-alien '|gpointer|)))
		(C-call "g_value_get_pointer" alien gvalue)
		alien))
	     ((int:= type (C-enum "G_TYPE_BOXED")) (unimplemented "a boxed"))
	     ((int:= type (C-enum "G_TYPE_PARAM")) (unimplemented "a param"))
	     ((int:= type (C-enum "G_TYPE_OBJECT"))
	      (let ((alien (make-alien '|GObject|)))
		(C-call "g_value_get_object" alien gvalue)
		alien))
	     (else
	      (error "Unexpected GFundamentalType:" type)))))
      (free gvalue)
      value)))

(define (gobject-set-properties gobject . property-list)
  (assert-glib-locked 'gobject-set-properties)
  (let* ((gobject-alien (gobject-alien gobject))
	 (gvalue (malloc (C-sizeof "GValue") '|GValue|))
	 (pspec (malloc (C-sizeof "GParamSpec") '|GParamSpec|))
	 (gtype (malloc (C-sizeof "GType") '|GType|))
	 (gclass (gobject-get-gclass gobject-alien))
	 (gclass-name (gclass-get-name gclass)))
    (let loop ((plist property-list))
      (cond ((null? plist) unspecific)
	    ((not (and (pair? plist) (pair? (cdr plist))))
	     (error "Odd length property list:" property-list))
	    (else
	     (let ((name (check-prop-name (car plist)))
		   (value (cadr plist)))
	       (C-call "g_object_class_find_property"
		       pspec gclass (string->utf8 name))
	       (if (alien-null? pspec)
		   (error "No property:" name gclass-name))
	       (let ((flags (C-> pspec "GParamSpec flags")))
		 (if (flag-set? flags (C-enum "G_PARAM_WRITABLE"))
		     (error "Property not writable:" name gclass-name))
		 (if (not (flag-set? flags (C-enum "G_PARAM_CONSTRUCT_ONLY")))
		     (error "Property not writable outside constructor:"
			    name gclass-name))
		 (C-call "G_PARAM_SPEC_VALUE_TYPE" gtype pspec)
		 (C-call "g_value_init" gvalue gtype)
		 ;;       g_value_set_* gvalue *
		 (let ((fundamental (C-call "G_TYPE_FUNDAMENTAL" gtype)))
		   (cond
		    ((int:= fundamental (C-enum "G_TYPE_CHAR"))
		     (C-call "g_value_set_schar"
			     gvalue (check-prop-char value name)))
		    ((int:= fundamental (C-enum "G_TYPE_UCHAR"))
		     (C-call "g_value_set_uchar"
			     gvalue (check-prop-uchar value name)))
		    ((int:= fundamental (C-enum "G_TYPE_INT"))
		     (C-call "g_value_set_int"
			     gvalue (check-prop-int value name)))
		    ((int:= fundamental (C-enum "G_TYPE_UINT"))
		     (C-call "g_value_set_uint"
			     gvalue (check-prop-uint value name)))
;		     (((C-enum "G_TYPE_LONG"))
;		      (C-call "g_value_set_long"
;			      gvalue (check-prop-long value name)))
;		     (((C-enum "G_TYPE_ULONG"))
;		      (C-call "g_value_set_ulong"
;			      gvalue (check-prop-ulong value name)))
		    ((int:= fundamental (C-enum "G_TYPE_FLOAT"))
		     (C-call "g_value_set_float"
			     gvalue (check-prop-flonum value name)))
		    ((int:= fundamental (C-enum "G_TYPE_DOUBLE"))
		     (C-call "g_value_set_double"
			     gvalue (check-prop-flonum value name)))
		    ((int:= fundamental (C-enum "G_TYPE_STRING"))
		     (C-call "g_value_set_string"
			     gvalue (check-prop-string value name)))
		    ((int:= fundamental (C-enum "G_TYPE_BOOLEAN"))
		     (C-call "g_value_set_boolean"
			     gvalue (check-prop-boolean value name)))
		    ((int:= fundamental (C-enum "G_TYPE_ENUM"))
		     (C-call "g_value_set_enum"
			     gvalue (check-prop-enum value name)))
		    ((int:= fundamental (C-enum "G_TYPE_FLAGS"))
		     (C-call "g_value_set_flags"
			     gvalue (check-prop-flags value name)))
		    ((int:= fundamental (C-enum "G_TYPE_OBJECT"))
		     (let* ((value-alien
			     (cond ((gobject? value) (gobject-alien value))
				   ((alien? value) value)
				   (else
				    (error "Property value not an object:"
					   value name gclass-name))))
			    (value-gtype
			     (gobject-get-gtype value-alien)))
		       (if (fix:zero? (C-call "g_value_type_compatible"
					      value-gtype gtype))
			   (error "Property value incompatible:"
				  value name gclass-name))
		       (C-call "g_value_set_object" gvalue value-alien)))
		    (else
		     (error "Property type unsupported:"
			    (or (C-enum "enum GFundamentalType" fundamental)
				fundamental)
			    name gclass-name))))
		 (C-call "g_object_set_property" gobject-alien
			 (string->utf8 name) gvalue)
		 (C-call "g_value_reset" gvalue)))
	     (loop (cddr plist)))))
    (free gtype)
    (free pspec)
    (free gvalue))
  unspecific)

(define (gobject-get-gclass alien)
  (assert-glib-locked 'gobject-get-gclass)
  (let ((ret (make-alien '|GObjectClass|)))
    (C-call "G_OBJECT_GET_CLASS" ret alien)
    ret))

(define (gclass-get-name gclass)
  (assert-glib-locked 'gclass-get-name)
  ;; GCLASS should be an alien of type GObjectClass.
  (let ((c* (make-alien '(* |gchar|))))
    (C-call "G_OBJECT_CLASS_NAME" c* gclass)
    (c-peek-cstring c*)))

(define (gobject-get-gtype gobject)
  (assert-glib-locked 'gobject-get-gtype)
  (let ((ret (make-alien '|GType|)))
    (C-call "G_OBJECT_TYPE" ret (gobject-alien gobject))
    ret))

(define (flag-set? fixnum mask)
  (not (fix:zero? (fix:and fixnum mask))))

(define (check-prop-name name)
  ;; Allows name to be a symbol OR string.
  (cond ((symbol? name) (symbol->string name))
	((string? name) name)
	(else (check-prop-name
	       (error "Invalid property name:" name)))))

(define (check-prop-value value property verb-phrase type-predicate)
  (if (type-predicate value) value
      (check-prop-value
       (error (string-append "Property value must " verb-phrase ":")
	      value property)
       property verb-phrase type-predicate)))

(define (check-prop-char value name)
  (check-prop-value value name "fit in a char"
		    (lambda (x) (and (fixnum? x)
				     (fix:<= -128 x) (fix:< x 128)))))

(define (check-prop-uchar value name)
  (check-prop-value value name "fit in an unsigned char"
		    (lambda (x) (and (fixnum? x) (fix:<= 0 x) (fix:< x 256)))))

(define (check-prop-int value name)
  (check-prop-value value name "fit in an int"
		    (lambda (x) (and (exact-integer? x)
				     (int:<= (expt -2 31) x)
				     (int:< x (expt 2 31))))))

(define (uint? x)
  (and (exact-integer? x) (int:<= 0 x) (int:< x (expt 2 32))))

(define (check-prop-uint value name)
  (check-prop-value value name "fit in an unsigned int" uint?))

#;(define (check-prop-long value name)
  (check-prop-value value name "fit in a long"
		    (lambda (x) (and (exact-integer? x)
				     (int:<= (expt -2 63) x)
				     (int:< x (expt 2 63))))))

#;(define (check-prop-ulong value name)
  (check-prop-value value name "fit in an unsigned long"
		    (lambda (x) (and (exact-integer? x)
				     (int:<= 0 x)
				     (int:< x (expt 2 64))))))

(define (check-prop-flonum value name)
  (check-prop-value value name "be a flonum" flo:flonum?))

(define (check-prop-string value name)
  (check-prop-value value name "be a string" string?))

(define (check-prop-boolean value name)
  (check-prop-value value name "be a boolean"
		    (lambda (x) (or (eq? #t x) (eq? #f x)))))

(define (check-prop-enum value name)
  (check-prop-value value name "be an enum" uint?))

(define (check-prop-flags value name)
  (check-prop-value value name "be a flagset" uint?))

(define (check-prop-gobject value name)
  (check-prop-value value name "be a gobject" gobject?))

;;; GQuarks

;;; No way (nor need) to GC.  Cache them here and toss cache when
;;; restored or reloaded.

(define gquark-from-string-cache (make-string-hash-table))

(define gquark-to-string-cache (make-key-weak-eqv-hash-table))

(define (gquark-from-string string)
  (assert-glib-locked 'gquark-from-string)
  (or (hash-table-ref/default gquark-from-string-cache string #f)
      (let ((gq (C-call "g_quark_from_string" (string->utf8 string))))
	(hash-table-set! gquark-from-string-cache string gq)
	(hash-table-set! gquark-to-string-cache gq string)
	gq)))

(define (gquark-to-string gquark)
  (or (hash-table-ref/default gquark-to-string-cache gquark #f)
      (error "Unknown GQuark:" gquark)))

(define (reset-quark-cache!)
  (set! gquark-from-string-cache (make-string-hash-table))
  (set! gquark-to-string-cache (make-key-weak-eqv-hash-table))
  unspecific)

(define (initialize-package!)
  (add-event-receiver! event:after-restore reset-quark-cache!)
  unspecific)

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))

(initialize-package!)